# Copyright (C) 2009-2010 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

from luci.lib.ClusterConf.Ip import Ip
from luci.lib.ClusterConf.Fs import Fs
from luci.lib.ClusterConf.Clusterfs import Clusterfs
from luci.lib.ClusterConf.Netfs import Netfs
from luci.lib.ClusterConf.NFSExport import NFSExport
from luci.lib.ClusterConf.NFSClient import NFSClient
from luci.lib.ClusterConf.NFSServer import NFSServer
from luci.lib.ClusterConf.Script import Script
from luci.lib.ClusterConf.Samba import Samba
from luci.lib.ClusterConf.Tomcat5 import Tomcat5
from luci.lib.ClusterConf.Tomcat6 import Tomcat6
from luci.lib.ClusterConf.Postgres8 import Postgres8
from luci.lib.ClusterConf.Apache import Apache
from luci.lib.ClusterConf.OpenLDAP import OpenLDAP
from luci.lib.ClusterConf.LVM import LVM
from luci.lib.ClusterConf.MySQL import MySQL
from luci.lib.ClusterConf.DRBD import DRBD
from luci.lib.ClusterConf.Vm import Vm
from luci.lib.ClusterConf.Named import Named
from luci.lib.ClusterConf.SAPDatabase import SAPDatabase
from luci.lib.ClusterConf.SAPInstance import SAPInstance
from luci.lib.ClusterConf.OracleDB import OracleDB
from luci.lib.ClusterConf.SybaseASE import SybaseASE
from luci.lib.ClusterConf.Service import Service
from luci.lib.ClusterConf.RefObject import RefObject

from xml.dom import minidom

from pylons.i18n import ugettext as _

import logging
log = logging.getLogger(__name__)

def get_fsid_list(model):
	obj_list = model.searchObjectTree('fs')
	obj_list.extend(model.searchObjectTree('clusterfs'))
	return map(lambda x: x.getAttribute('fsid') and int(x.getAttribute('fsid')) or 0, obj_list)

def fsid_is_unique(model, fsid):
	fsid_list = get_fsid_list(model)
	return fsid not in fsid_list

def generate_fsid(model, name):
	import binascii
	from random import random
	fsid_list = get_fsid_list(model)

	fsid = binascii.crc32(name) & 0xffff
	dupe = fsid in fsid_list
	while dupe is True:
		fsid = (fsid + random.randrange(1, 0xfffe)) & 0xffff
		dupe = fsid in fsid_list
	return fsid

def getResourceForEdit(model, name):
	resPtr = model.getResourcesPtr()
	resources = resPtr.getChildren()

	for res in resources:
		if res.getName() == name:
			resPtr.removeChild(res)
			return res
	raise KeyError, name

def config_resource(params, res, rname, **kw):
	errors = list()

	for i in params:
		# parameter (str), short desc (str), required (bool), default (str or None)
		name = i[0]
		val = kw.get(name)
		required = i[2]
		default_val = i[3]

		try:
			val = val.strip()
			if not val:
				val = None
		except:
			val = None

		if not val:
			if required is True:
				if default_val is None:
					errors.append(_('No value for required attribute "%s" was given for %s resource %s') % (name, i[1], rname))
				else:
					res.addAttribute(name, default_val)
			else:
				try:
					res.removeAttribute(name)
				except:
					pass
		else:
			res.addAttribute(name, val)
	return errors

def addIp(res, rname, model, **kw):
	params = (
		('address', _('IP address'), True, None),
		('monitor_link', _('Monitor Link'), False, None),
		('sleeptime', _('Sleep time'), False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addFs(res, rname, model, **kw):
	params = (
		('mountpoint', _('Mountpoint'), True, None),
		('device', _('Device, label, or UUID'), True, None),
		('fstype', '', False, None),
		('force_unmount', '', False, None),
		('quick_status', '', False, None),
		('self_fence', '', False, None),
		('fsid', '', False, None),
		('force_fsck', '', False, None),
		('options', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	fsid = res.getAttribute('fsid')
	if not fsid:
		fsid = str(generate_fsid(model, rname))
		res.addAttribute('fsid', fsid)
	else:
		if not fsid_is_unique(model, int(fsid)):
			errors.append(_('fsid for resource %s is not unique') % rname)
	return errors

def addClusterfs(res, rname, model, **kw):
	params = (
		('mountpoint', '', True, None),
		('device', '', True, None),
		('fstype', '', False, None),
		('force_unmount', '', False, None),
		('self_fence', '', False, None),
		('fsid', '', False, None),
		('options', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	fsid = res.getAttribute('fsid')
	if not fsid:
		fsid = str(generate_fsid(model, rname))
		res.addAttribute('fsid', fsid)
	else:
		if not fsid_is_unique(model, int(fsid)):
			errors.append(_('fsid for resource %s is not unique') % rname)
	return errors

def addNetfs(res, rname, model, **kw):
	params = (
		('mountpoint', _('Mountpoint'), True, None),
		('host', _('NFS/CIFS server host'), True, None),
		('export', _('Export name'), True, None),
		('options', '', False, None),
		('no_unmount', '', False, None),
		('force_unmount', '', False, None),
		('fstype', '', False, None)
	)

	errors = config_resource(params, res, rname, **kw)
	return errors

def addVM(res, rname, model, **kw):
	params = (
		('migration_mapping', '', False, None),
		('xmlfile', '', False, None),
		('migrate', '', False, None),
		('path', '', False, None),
		('snapshot', '', False, None),
		('hypervisor_uri', '', False, None),
		('migration_uri', '', False, None),
		('status_program', '', False, None)
	)

	errors = config_resource(params, res, rname, **kw)
	return errors

def addNFSClient(res, rname, model, **kw):
	params = (
		('target', _('Hosts to export to'), True, None),
		('options', '', False, None),
		('allow_recover', '', False, None)
	)

	errors = config_resource(params, res, rname, **kw)
	return errors

def addNFSExport(res, rname, model, **kw):
	# Only the name is used
	return []

def addNFSServer(res, rname, model, **kw):
	params = (
		('nfspath', '', False, None),
	)

	errors = config_resource(params, res, rname, **kw)
	return errors

def addScript(res, rname, model, **kw):
	params = (
		('file', _('Path to script'), True, None),
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addSamba(res, rname, model, **kw):
	params = (
		('config_file', _('Path to configuration file'), False, None),
		('smbd_options', '', False, None),
		('nmbd_options', '', False, None),
		('shutdown_wait', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addApache(res, rname, model, **kw):
	params = (
		('config_file', _('Path to configuration file'), False, None),
		('server_root', '', False, None),
		('httpd_options', '', False, None),
		('shutdown_wait', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addMySQL(res, rname, model, **kw):
	params = (
		('config_file', _('Path to configuration file'), False, None),
		('listen_address', '', False, None),
		('mysqld_options', '', False, None),
		('startup_wait', '', False, None),
		('shutdown_wait', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addOpenLDAP(res, rname, model, **kw):
	params = (
		('config_file', _('Path to configuration file'), False, None),
		('url_list', '', False, None),
		('slapd_options', '', False, None),
		('shutdown_wait', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addPostgres8(res, rname, model, **kw):
	params = (
		('config_file', _('Path to configuration file'), False, None),
		('postmaster_user', '', False, None),
		('postmaster_options', '', False, None),
		('shutdown_wait', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addTomcat5(res, rname, model, **kw):
	params = (
		('config_file', _('Path to configuration file'), False, None),
		('tomcat_user', '', False, None),
		('catalina_options', '', False, None),
		('catalina_base', '', False, None),
		('shutdown_wait', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addTomcat6(res, rname, model, **kw):
	params = (
		('config_file', _('Path to configuration file'), False, None),
		# XXX really not supported?
		# ('tomcat_user', '', False, None),
		('shutdown_wait', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addLVM(res, rname, model, **kw):
	params = (
		('vg_name', _('Volume group name'), True, None),
		('lv_name', '', False, None),
		('self_fence', '', False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addSAPDatabase(res, rname, model, **kw):
	params = (
		('DBTYPE', _('Database type'), True, None),
		('DIR_EXECUTABLE', '', False, None),
		('NETSERVICENAME', '', False, None),
		('DBJ2EE_ONLY', '', False, None),
		('JAVA_HOME', '', False, None),
		('STRICT_MONITORING', '', False, None),
		('AUTOMATIC_RECOVER', '', False, None),
		('DIR_BOOTSTRAP', '', False, None),
		('DIR_SECSTORE', '', False, None),
		('DB_JARS', '', False, None),
		('PRE_START_USEREXIT', '', False, None),
		('POST_START_USEREXIT', '', False, None),
		('PRE_STOP_USEREXIT', '', False, None),
		('POST_STOP_USEREXIT', '', False, None),
	)
	errors = config_resource(params, res, rname, **kw)
	res.removeAttribute('name')
	res.addAttribute('SID', rname)

	dbtype = res.getAttribute('DBTYPE')
	if not dbtype:
		dbtype = 'ORA'
		res.addAttribute('DBTYPE', 'ORA')
	if not dbtype.upper() in ( 'ORA', 'DB6', 'ADA' ):
		errors.append(_('Invalid database type: %s') % dbtype)
	return errors

def addSAPInstance(res, rname, model, **kw):
	params = (
		('DIR_EXECUTABLE', '', False, None),
		('DIR_PROFILE', '', False, None),
		('START_PROFILE', '', False, None),
		('START_WAITTIME', '', False, None),
		('AUTOMATIC_RECOVER', '', False, None),
		('PRE_START_USEREXIT', '', False, None),
		('POST_START_USEREXIT', '', False, None),
		('PRE_STOP_USEREXIT', '', False, None),
		('POST_STOP_USEREXIT', '', False, None)
	)

	errors = config_resource(params, res, rname, **kw)
	res.removeAttribute('name')
	res.addAttribute('InstanceName', rname)
	return errors

def addSybaseASE(res, rname, model, **kw):
	params = (
		('sybase_home', _('SYBASE home directory'), True, None),
		('sybase_ase', _('SYBASE_ASE directory name'), True, None),
		('sybase_ocs', _('SYBASE_OCS directory name'), True, None),
		('server_name', _('ASE server name'), True, None),
		('login_file', _('Login file'), True, None),
		('interfaces_file', _('Interfaces file'), True, None),
		('sybase_user', _('Sybase user'), True, 'sybase'),
		('shutdown_timeout', _('Shutdown timeout value'), True, None),
		('start_timeout', _('Start timeout value'), True, None),
		('deep_probe_timeout', _('Deep probe timeout value'), True, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addOracleDB(res, rname, model, **kw):
	params = (
		('user', _('Oracle User Name'), True, None),
		('home', _('Oracle Home Directory'), True, None),
		('listener_name', '', False, None),
		('oracletype', _('Oracle Installation Type'), True, '10g'),
		('vhost', _('Virtual Hostname'), False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	dbtype = res.getAttribute('oracletype')
	if not dbtype.lower() in ('base', 'base-em', 'ias', '10g', '10g-ias'):
		errors.append(_('Invalid Oracle database type "%s"') % dbtype)
	if dbtype:
		res.removeAttribute('oracletype')
		res.addAttribute('type', dbtype)
	return errors

def addNamed(res, rname, model, **kw):
	params = (
		('config_file', _('named config file'), False, None),
		('named_working_dir', _('named working directory'), False, None),
		('named_options', _('Other command line options'), False, None),
		('shutdown_wait', _('Shutdown Wait Time'), False, None)
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

def addDRBD(res, rname, model, **kw):
	params = (
		('config_file', _('drbd config file'), False, None),
		('resource', _('drbd resource name'), True, None),
	)
	errors = config_resource(params, res, rname, **kw)
	return errors

resource_table = {
	'ip':			( addIp,			Ip			),
	'fs':			( addFs,			Fs			),
	'clusterfs':	( addClusterfs,		Clusterfs	),
	'netfs':		( addNetfs,			Netfs		),
	'nfsexport':	( addNFSExport,		NFSExport	),
	'nfsclient':	( addNFSClient,		NFSClient	),
	'nfsserver':	( addNFSServer,		NFSServer	),
	'script':		( addScript,		Script		),
	'smb':			( addSamba,			Samba		),
	'tomcat-5':		( addTomcat5,		Tomcat5		),
	'tomcat-6':		( addTomcat6,		Tomcat6		),
	'postgres-8':	( addPostgres8,		Postgres8	),
	'apache':		( addApache,		Apache		),
	'openldap':		( addOpenLDAP,		OpenLDAP	),
	'lvm':			( addLVM,			LVM			),
	'mysql':		( addMySQL,			MySQL		),
	'SAPDatabase':	( addSAPDatabase,	SAPDatabase	),
	'SAPInstance':	( addSAPInstance,	SAPInstance	),
	'oracledb':		( addOracleDB,		OracleDB	),
	'ASEHAagent':	( addSybaseASE,		SybaseASE	),
	'named':		( addNamed,			Named		),
	'drbd':			( addDRBD,			DRBD		),
	'vm':			( addVM,			Vm			),
}

def create_resource(res_type, model, **kw):
	errors = list()
	if not res_type:
		raise Exception, _('No resource type was given')

	if not resource_table.has_key(res_type):
		raise Exception, _('Unknown resource type: "%s"') % res_type

	res = None
	if kw.has_key('edit'):
		try:
			if not kw.has_key('oldname'):
				raise Exception, _('Cannot find this resource\'s original name.')

			oldname = kw['oldname'].strip()
			if not oldname:
				raise Exception, _('Cannot find this resource\'s original name.')

			res = getResourceForEdit(model, oldname)
			if not res:
				raise Exception, _('Cluster model object for "%s" not found') % oldname
		except Exception, e:
			log.exception('Error adding resource: %s' % str(e))
			raise Exception, _('No resource named "%s" exists.') % oldname
	else:
		res = resource_table[res_type][1]()

	if res_type == 'ip':
		rname = kw['address']
	elif res_type == 'vm':
		rname = ''
	else:
		if not kw.has_key('resourcename') or not kw['resourcename'].strip():
			raise Exception, _('All resources must have a unique name.')
		rname = kw['resourcename'].strip()
		res.addAttribute('name', rname)

	if not kw.has_key('edit'):
		try:
			dummy = getResourceForEdit(model, rname)
			if dummy:
				raise Exception, _('A resource named "%s" already exists.') % rname
		except:
			pass

	errors = resource_table[res_type][0](res, rname, model, **kw)

	if len(errors) > 0:
		raise Exception, ', '.join(errors)

	return res

def validate_clusvc_form(model, **kw):
	errors = list()

	form_xml = kw.get('form_xml')
	if not form_xml:
		form_xml = ''

	root_elem = kw.get('root_elem')
	if not root_elem:
		root_elem = 'svc_root'

	forms = []
	if form_xml.strip():
		try:
			doc = minidom.parseString(form_xml)
			forms = doc.getElementsByTagName('form')
			if len(forms) < 1:
				raise Exception, 'invalid XML'
		except:
			log.exception('Error parsing service XML: "%s"' % form_xml)
			return (False, { 'errors': [ 'The resource data submitted for this service is not properly formed' ]})

	is_vm = False
	form_hash = {}
	form_hash[root_elem] = { 'form': None, 'kids': [] }
	for i in forms:
		form_id = i.getAttribute('id')
		form_parent = i.getAttribute('parent')
		if not form_id or not form_parent:
			continue
		ielems = i.getElementsByTagName('input')
		if not ielems or len(ielems) < 1:
			continue
		if not form_id in form_hash:
			form_hash[form_id] = {'form': i, 'kids': []}
		elif not form_hash[form_id]['form']:
			form_hash[form_id]['form'] = i
		if not form_parent in form_hash:
			form_hash[form_parent] = {'form': None, 'kids': []}
		form_hash[form_parent]['kids'].append(form_id)
		dummy_form = {}

		for i in ielems:
			try:
				input_type = str(i.getAttribute('type'))
			except:
				log.exception('No input type')
				continue

			if not input_type or input_type == 'button':
				continue

			try:
				dummy_form[str(i.getAttribute('name'))] = str(i.getAttribute('value'))
			except:
				log.exception('Error parsing form XML values')

		try:
			res_type = dummy_form['type'].strip()
			if not res_type:
				raise Exception, 'no resource type'
			if res_type == 'vm':
				is_vm = True
		except Exception, e:
			log.exception('no resource type')
			return (False, { 'errors': [ 'No resource type was specified' ]})

		try:
			if res_type == 'ip':
				mask = dummy_form.get('address_mask')
				if mask:
					dummy_form['resourcename'] = '%s/%s' % (dummy_form['address_nominal'], dummy_form['address_mask'])
				else:
					dummy_form['resourcename'] = dummy_form['address_nominal']
				dummy_form['address'] = dummy_form['resourcename']
		except:
			log.exception('no ipaddr')
			return (False, { 'errors': [ 'No IP address was given' ]})

		try:
			if dummy_form.has_key('global'):
				newRes = model.getResourceByName(dummy_form['resourcename'])
				resObj = RefObject(newRes)
				resObj.setRef(newRes.getName())
			else:
				resObj = create_resource(res_type, model, **dummy_form)
		except Exception, e:
			resObj = None
			errors.append(_('Error adding resource type %s: %s') % (res_type, str(e)))
			log.exception('Error validating %s resource' % res_type)

		if not resObj:
			return (False, {'errors': errors})

		if dummy_form.has_key('__independent_subtree'):
			resObj.addAttribute('__independent_subtree', '1')
		else:
			resObj.removeAttribute('__independent_subtree')
		form_hash[form_id]['obj'] = resObj

	if len(errors) > 0:
		return (False, {'errors': errors})

	fdom = kw.get('domain')

	recovery = kw.get('recovery')
	if not recovery.lower() in ('restart', 'relocate', 'disable'):
		errors.append('You entered an invalid recovery option: "%s"' % recovery)

	max_restarts = None
	restart_expire_time = None
	if recovery == 'restart':
		max_restarts = kw.get('max_restarts')
		if max_restarts:
			try:
				max_restarts = int(max_restarts)
				if max_restarts < 0:
					raise ValueError
			except Exception, e:
				errors.append('Maximum restarts must be a number greater than or equal to 0')
				max_restarts = None

		restart_expire_time = kw.get('restart_expire_time')
		if restart_expire_time:
			try:
				restart_expire_time = int(restart_expire_time)
				if restart_expire_time < 0:
					raise ValueError
			except Exception, e:
				errors.append('Restart expire time must be a number greater than or equal to 0')
				restart_expire_time = None

	service_name = kw.get('svc_name')
	if not service_name:
		errors.append('No service name was given')
	old_name = kw.get('old_name')

	autostart = '1'
	as_val = kw.get('autostart')
	if not as_val:
		autostart = '0'

	exclusive = '0'
	ex_val = kw.get('exclusive')
	if not ex_val:
		exclusive = '0'
	else:
		exclusive = '1'

	action = kw.get('action')
	if action is None:
		return (False, {'errors': [ _('No action was given for service "%s"') % service_name ] })

	action = action.lower()
	if action == 'edit':
		cur_service = model.retrieveServiceByName(old_name)
		if cur_service is None:
			return (False, {'errors': [ _('The service "%s" could not be found for editing') % service_name ]})
		model.deleteService(service_name)
	elif action == 'create':
		cur_service = model.retrieveServiceByName(service_name)
		if cur_service is not None:
			return (False, {'errors': [ _('A cluster service with the name "%s" already exists') % service_name ]})
	else:
		return (False, {'errors': [ _('An unknown action "%s" was specified') % action ]})

	def buildSvcTree(parent, child_id_list):
		for i in child_id_list:
			try:
				child = form_hash[i]['obj']
				if not child:
					raise Exception, 'No object for %s' % i
			except Exception, e:
				log.exception('Error building service tree')
				continue
			parent.addChild(child)
			if 'kids' in form_hash[i]:
				buildSvcTree(child, form_hash[i]['kids'])

	new_service = Service()
	new_service.addAttribute('name', service_name)
	if fdom:
		new_service.addAttribute('domain', fdom)

	if recovery:
		new_service.addAttribute('recovery', recovery)

	if max_restarts is not None:
		new_service.addAttribute('max_restarts', str(max_restarts))

	if restart_expire_time is not None:
		new_service.addAttribute('restart_expire_time', str(restart_expire_time))

	new_service.addAttribute('exclusive', str(exclusive))
	if autostart is not None:
		new_service.addAttribute('autostart', str(autostart))

	buildSvcTree(new_service, form_hash[root_elem]['kids'])

	if is_vm is True:
		new_vm = Vm()
		svc_children = new_service.getChildren()
		if len(svc_children) != 1:
			errors.append('VMs may have no children and may not be children of other resources')
		else:
			vm_res = svc_children[0]
			vm_attr = vm_res.getAttributes()
			vm_attr.update(new_service.getAttributes())
			for k, v in vm_attr.items():
				new_vm.addAttribute(k, v)
		model.resourcemanager_ptr.addChild(new_vm)
	else:
		model.resourcemanager_ptr.addChild(new_service)
	model.setModified(True)
	return (len(errors) == 0, {'errors': errors})

def validate_resource_form(model, **kw):
	errors = []

	try:
		res_type = kw.get('type')
		res = create_resource(res_type, model, **kw)
		model.getResourcesPtr().addChild(res)
	except Exception, e:
		errors.append(_('Error adding resource type %s: %s') % (res_type, str(e)))
		return (False, {'errors': errors})
	return (True, {})
